import argparse
from typing import IO, Iterable
import json

from multiprocessing import Pool
from nanoepitools.haplotype_tools import extract_read_haplotype_assignment


def extract_haplotype_ids(
    bam: IO, output: IO, chroms: Iterable[str], include_unphased: bool
):
    print("Reading bam file")
    if len(chroms) == 1 and chroms[0] == "all":
        chroms = None
    ps_collection = extract_read_haplotype_assignment(
        bam, chroms=chroms, return_unphased=include_unphased
    )
    print("Writing output")

    out_line = "{read_name}\t{hp}\t{ps}\n"

    with open(output, "w") as out_f:
        if include_unphased:
            for k, reads in ps_collection.unphased.items():
                for read in reads:
                    out_f.write(out_line.format(read_name=read.read_name, hp=-1, ps=-1))

        for ps_id, ps in ps_collection.phased.items():
            for chrom_reads in ps.read_mappings.values():
                for read in chrom_reads:
                    out_f.write(
                        out_line.format(
                            read_name=read.read_name, hp=read.haplotype, ps=ps_id
                        )
                    )



import argparse
from typing import IO, Iterable

from multiprocessing import Pool
import pandas as pd
from meth5.meth5_wrapper import MetH5File


def read_readgroups(readgroups_file: IO):
    """
    Reads file that assigns read to read groups (such as haplotypes,
    samples, clusters, etc)
    :param readgroups_file: path to the tab-separated file
    :return: pandas dataframe with columns "read_name", "group" and "group_set"
    """
    # Loading
    try:
        read_groups = pd.read_csv(
            readgroups_file,
            sep="\t",
            header=0,
            dtype={"read_name": str, "group": int, "group_set": "category"},
        )
    except Exception as e:
        logging.error("Unable to read read groups file", e)
        raise e

    # Validation
    if len(read_groups.columns) == 2:
        should_colnames = ["read_name", "group"]
    elif len(read_groups.columns) == 3:
        should_colnames = ["read_name", "group", "group_set"]
    else:
        logging.error(
            "Invalid number of columns in read groups file (should be 2 or 3)"
        )
        sys.exit(1)

    if not all([col in read_groups.columns for col in should_colnames]):
        logging.error(
            "Invalid column names in read groups file (should be %s)"
            % should_colnames.join(", ")
        )
        sys.exit(1)

    # Finished validation, now add group_set column if not present
    if "group_set" not in read_groups.columns:
        read_groups["group_set"] = 1
    return read_groups


def chunked_segmentation(
    h5file: IO,
    chunksize: int,
    chunks: Iterable[int],
    readgroups_file: IO,
    workers: int,
    include_nogroup: bool,
):
    pool = Pool(workers)
    print(h5file)

    print("Reading read groups")
    read_groups = read_readgroups(readgroups_file)
    if not include_nogroup:
        read_groups = read_groups.loc[read_groups["group"] != -1]

    read_groups = read_groups.groupby("group_set")



from typing import Tuple

import numpy as np
import scipy.optimize

from nanoepiseg.emissions import EmissionLikelihoodFunction


def arraylogexpsum(x):
    ret = x[0]
    for i in range(1, len(x)):
        ret = logaddexp(ret, x[i])
    return ret  # if ret > -256 else -512


def logaddexp(a, b):
    ret = np.logaddexp(a, b)
    return ret


class SegmentationHMM:
    def __init__(self, max_segments: int, t_stay: float, t_move: float,
                 e_fn: EmissionLikelihoodFunction, seg_penalty: float = 0,
                 eps: np.float64 = np.exp(-512)):
        self.eps = eps
        self.e_fn = e_fn

        self.num_segments = max_segments

        if t_stay + t_move + seg_penalty > 1:
            raise ValueError('t_stay + t_move + seg_penalty may not exceed 1')

        self.seg_penalty = np.array(
            [(seg_penalty * i / max_segments) for i in range(max_segments)],
            dtype=np.float64)
        self.t_move = np.array(
            [t_move - self.seg_penalty[i] for i in range(max_segments)],
            dtype=np.float64)
        self.t_stay = np.array([t_stay for i in range(max_segments)],
                               dtype=np.float64)
        self.t_end = np.array(
            [1 - self.t_move[i] - self.t_stay[i] for i in range(max_segments)],
            dtype=np.float64)
        self.t_move = np.log(self.t_move + eps)
        self.t_stay = np.log(self.t_stay + eps)
        self.t_end = np.log(self.t_end + eps)

    def t_fn(self, i, j):
        if i == j:
            return self.t_stay[i]
        if i == (j - 1):
            # Probability to move to the next state 
            return self.t_move[i]  # + sim_penalty
        if j == (self.num_segments - 1):
            # Probability to go the last segment
            return self.t_end[i]

        raise RuntimeError(
            'Transition %d to %d is not a valid transition in segmentation '
            'HMM ' % (i, j))

    def forward(self, observations, obs_c):
        e_fn = self.e_fn.likelihood
        M = self.num_segments
        R = observations.shape[0]
        N = observations.shape[1]
        F = np.zeros((N, M), dtype=np.float) + self.eps
        F[0, 0] = 1 - F[0, :].sum() - self.eps
        F = np.log(F)
        start_prob = np.zeros(M) + self.eps
        start_prob[0] = 1
        start_prob = np.log(start_prob)

        for k in range(N):
            o = observations[:, k]
            for i in range(M):
                e = e_fn(i, o, obs_c)

                if k == 0:
                    F[k, i] = e + start_prob[i]
                    continue

                    # Stay probability
                F[k, i] = e + F[k - 1, i] + self.t_fn(i, i)

                # Move probabilty
                if i > 0:
                    F[k, i] = logaddexp(F[k, i],
                                        e + F[k - 1, i - 1] + self.t_fn(i - 1,
                                                                        i))

                # End probability
                if i == M - 1:
                    # if end state we could have come from anywhere to the
                    # end state:
                    for j in range(
                            M - 2):  # exclude last 2 because those were already
                        # handled above
                        F[k, i] = logaddexp(F[k, i],
                                            e + F[k - 1, j] + self.t_fn(j, i))
        evidence = F[-1, -1]
        return F, evidence

    def backward(self, observations, obs_c):
        e_fn = self.e_fn.likelihood
        R = observations.shape[0]
        M = self.num_segments
        N = observations.shape[1]
        B = np.zeros((N, M), dtype=np.float64) + self.eps
        B[-1, -1] = 1
        B = np.log(B)

        for k in range(N - 1, 0, -1):
            o = observations[:, k]
            k = k - 1
            for i in range(M):
                e_stay = e_fn(i, o, obs_c)

                if i == M - 1:
                    # If i is end state, we can only stay
                    B[k, i] = e_stay + B[k + 1, i] + self.t_fn(i, i)
                else:
                    e_move = e_fn(i + 1, o, obs_c)
                    # Move and stay probability
                    B[k, i] = logaddexp(B[k + 1, i] + self.t_fn(i, i) + e_stay,
                                        B[k + 1, i + 1] + self.t_fn(i,
                                                                    i + 1) +
                                        e_move)
                    if i < M - 2:
                        # End probability only if i<M-2 because otherwise it
                        # was covered by move or stay
                        e_end = e_fn(M - 1, o, obs_c)
                        B[k, i] = logaddexp(B[k, i],
                                            B[k + 1, M - 1] + self.t_fn(i,
                                                                        M -
                                                                        1) +
                                            e_end)

        o = observations[:, 0]
        evidence = B[0, 0] + e_fn(0, o, obs_c)
        return B, evidence

    def viterbi(self, observations, obs_c):
        e_fn = self.e_fn.likelihood
        M = self.num_segments
        N = observations.shape[1]

        V = np.zeros((N, M), dtype=np.float64) + self.eps
        V[0, 0] = 1
        V = np.log(V)
        P = np.zeros((N, M), dtype=np.int32)

        start_prob = np.zeros(M) + self.eps
        start_prob[0] = 1
        start_prob = np.log(start_prob)

        for k in range(0, N - 1):
            o = observations[:, k]
            for i in range(M):
                e = e_fn(i, o, obs_c)

                if k == 0:
                    V[k, i] = np.max(e + start_prob[i])
                    continue

                p = np.zeros(M) - np.inf

                p[i] = V[k - 1, i] + self.t_fn(i, i)

                if i > 0:
                    p[i - 1] = V[k - 1, i - 1] + self.t_fn(i - 1, i)

                if i == M - 1:
                    # last two have been covered by stay and move
                    for j in range(M - 2):
                        p[j] = V[k - 1, j] + self.t_fn(j, i)
                p = e + p

                V[k, i] = np.max(p)
                P[k, i] = np.argmax(p)
            # Rescaling prevents underflow
            V[k, :] = V[k, :] - arraylogexpsum(V[k, :])
        V[-1, :] = np.log(self.eps)
        V[-1, -1] = np.max(V[-2, :])
        P[-1, -1] = np.argmax(V[-2, :])
        X = np.zeros(N, dtype=np.int32)
        Z = np.zeros(N, dtype=np.float32)
        X[N - 1] = M - 1
        Z[N - 1] = 0

        for k in range(N - 2, -1, -1):
            X[k] = P[k + 1, X[k + 1]]
            Z[k] = V[k + 1, X[k + 1]]

        return X, Z

    def MAP(self, posterior):
        M = self.num_segments
        N = posterior.shape[0]

        V = np.zeros((N, M), dtype=np.float) + self.eps
        V[0, 0] = 1
        V = np.log(V)
        P = np.zeros((N, M), dtype=np.int32)

        start_prob = np.zeros(M) + self.eps
        start_prob[0] = 1
        start_prob = np.log(start_prob)

        for k in range(0, N - 1):
            for i in range(M):
                e = posterior[k, i]

                if k == 0:
                    V[k, i] = np.max(e + start_prob[i])
                    continue

                p = np.zeros(M) - np.inf

                p[i] = V[k - 1, i] + self.t_fn(i, i)

                if i > 0:
                    p[i - 1] = V[k - 1, i - 1] + self.t_fn(i - 1, i)

                if i == M - 1:
                    for j in range(
                            M - 2):  # last two have been covered by stay and
                        # move
                        p[j] = V[k - 1, j] + self.t_fn(j, i)
                p = e + p

                V[k, i] = np.max(p)
                P[k, i] = np.argmax(p)
            # Rescaling prevents underflow
            V[k, :] = V[k, :] - arraylogexpsum(V[k, :])
        V[-1, :] = np.log(self.eps)
        V[-1, -1] = np.max(V[-2, :])
        P[-1, -1] = np.argmax(V[-2, :])
        X = np.zeros(N, dtype=np.int32)
        Z = np.zeros(N, dtype=np.float32)
        X[N - 1] = M - 1
        Z[N - 1] = 0

        for k in range(N - 2, -1, -1):
            X[k] = P[k + 1, X[k + 1]]
            Z[k] = V[k + 1, X[k + 1]]

        return X, Z

    def baum_welch(self, observations: np.ndarray, tol: float = np.exp(-4),
                   it_hook=None, samples: np.ndarray = None,
                   verbose: bool = False) -> Tuple[np.ndarray, np.ndarray]:
        """
        Run the baum_welch algorithm, an expectation maximization algorithm,
        to find a segmentation of the methylation signal.

        Note that this algorithm is rather memory expensive. It will take
        O(CM) memory where C is the number of samples and M the maximum number
        of segments. If no samples are provided, C is equal to the number of
        reads, meaning the memory requirement grows with the read coverage.

        :param observations: a numpy array of shape RxN, where R is the
        number of reads and N is the number of gennomic positions (or CpG
        sites). The values need to be in the range (0,1) and are methylation
        predictions for the individual CpG sites. In order to speed up
        computation, missing predictions can be labeled with the value -1.
        This should lead to the same result as setting it to the value 0.5,
        but reduces the number of computations required significantly.
        :param tol: The absolute maximum difference in a parameter value that
        determines convergence. If the difference is below tol, the algorithm
        aborts
        :param it_hook: A function hook that will be called after each
        iteration. Takes the same parameters as the return value of this
        function
        :param samples: A 1-dimensional numpy array of length R, assigns each
        read to a sample id. Sample ids must be integer, and must start from
        0 and have no gaps.
        :return: tuple with estimated parameters and posteriors. Estimated
        paramater type depends on the given emission probability class.
        Posterior is of shape NxM and gives the posterior probability of each
        genomic site n being in each segment m
        """
        # Initial guess of parameters
        R = observations.shape[0]

        if any((observations != -1).sum(axis=0) == 0):
            raise ValueError('Observations must not include reads with no '
                             'observations')
        if any((observations != -1).sum(axis=1) == 0):
            raise ValueError('Observations must not include sites with no '
                             'observations')

        if samples is None:
            # No samples, then we use the identity
            self.obs_c = np.arange(R)
        else:
            self.obs_c = samples

        C = len(set(self.obs_c))

        for it in range(100):
            F, f_evidence = self.forward(observations, self.obs_c)
            B, b_evidence = self.backward(observations, self.obs_c)
            # Sanity check: fwd and bwd algorithm should return same evidence
            if np.abs(f_evidence - b_evidence) > 10e-6:
                print("WARNING: forward evidence %f does not equal backward "
                      "evidence %f." % (f_evidence, b_evidence))

            posterior = F + B - b_evidence

            # Maximize
            segment_p_new = {}

            for c in range(C):
                old_params = self.e_fn.get_cluster_params(c)
                to_minimize = self.e_fn.minimization_objective(
                    observations[self.obs_c == c], np.exp(posterior))
                bounds = self.e_fn.get_param_bounds()

                estimated_p = scipy.optimize.minimize(to_minimize, old_params,
                                                          method='SLSQP',
                                                          bounds=bounds).x

                segment_p_new[c] = np.log(estimated_p)

            diff = self.e_fn.update_params(segment_p_new)

            segment_p = segment_p_new

            if it_hook is not None:
                it_hook(segment_p, posterior)

            if verbose:
                print("Iteration %d, parameter difference: %f" % (it, diff))
            if diff < tol:
                break

        return segment_p, posterior



import sys
import argparse
from pathlib import Path
from typing import IO, List, Tuple, Dict

from multiprocessing import Pool
from meth5.meth5_wrapper import (
    MetH5File,
    MethlyationValuesContainer,
    ChromosomeContainer,
    create_sparse_matrix_from_samples,
)

'''

def segmentation_worker(window_met, rn_hp_dict, n_segments):
    dense_met_llr = np.array(window_met.met_matrix.todense())
    dense_met = dense_met_llr.copy()
    met_prob = net_m.llr_to_p(dense_met[dense_met != 0])
    # We need to ensure some uncertainty, since otherwise likelihoods explode
    met_prob = np.clip(met_prob, np.exp(-16), 1 - np.exp(-16))
    dense_met[dense_met != 0] = met_prob
    dense_met[dense_met == 0] = -1

    if len(dense_met.shape) != 2:
        # Not sure if this ever happens - but just in case the array
        # collapses if too much has been filtered out
        raise ValueError("Input matrix has wrong dimensions")
    if dense_met.shape[0] == 0 or dense_met.shape[1] < 10:
        raise ValueError("Empty input matrix")

    # Since haplotypes might be arbitrary integers, but the
    # segmentation HMM needs integers starting from 0, we translate
    # haplotype ids here to integers starting from 0
    window_hp = [rn_hp_dict[read] for read in window_met.read_names]
    hp_index_dict = {hp: i for i, hp in enumerate(sorted(list(set(window_hp))))}
    window_hp = np.array([hp_index_dict[hp] for hp in window_hp])

    num_samples = len(set(window_hp))

    use_samples_in_hmm = True
    if use_samples_in_hmm:
        emission_lik = BernoulliPosterior(num_samples, n_segments, prior_a=None)
        hmm = SegmentationHMM(n_segments, 0.1, 0.8, emission_lik)
        hmm.baum_welch(dense_met, samples=window_hp)
    else:
        emission_lik = BernoulliPosterior(len(window_hp), n_segments, prior_a=None)
        hmm = SegmentationHMM(n_segments, 0.1, 0.8, emission_lik)
        hmm.baum_welch(dense_met)
    segments_p = emission_lik.segment_p
    segmentation, _ = hmm.viterbi(dense_met, window_hp)
    segmentation = nes_pp.cleanup_segmentation(segments_p, segmentation)
    return segmentation, segments_p, hp_index_dict, dense_met_llr, window_hp


class SegmentationCallback:
    def __init__(
        self,
        sw: SegmentsWriterCSV,
        pa: PlotArchiver,
        window_met,
        ps_id: int,
        chrom: str,
    ):
        self.sw = sw
        self.pa = pa
        self.window_met = window_met
        self.ps_id = ps_id
        self.chrom = chrom

    def segmentation_callback(self, result):
        segmentation, segments_p, hp_index_dict, dense_met_llr, window_hp = result
        print("plotting")
        plt.figure(figsize=(10, 8))
        title_format = "{chrom}:{start}-{end}, Phase Set {ps}"
        start = self.window_met.get_genomic_region()[0]
        end = self.window_met.get_genomic_region()[1]
        plt.title(
            title_format.format(chrom=self.chrom, start=start, end=end, ps=self.ps_id)
        )

        """We need to translate haplotype id (from whatshap: 1 and 2) into what
        the segmentation algorithm knows as the sample index (0 and 1 in
        this case). This also needs to consider that sometimes there is
        only a single haplotype in a window or even the entire phase set,
        which is why we need to check this when constructing the color
        dict"""
        plot_hpindex_colors = {
            hp_index_dict[x]: plot_hp_colors[x]
            for x in plot_hp_colors.keys()
            if x in hp_index_dict.keys()
        }

        net_pm.plot_met_profile(
            dense_met_llr,
            window_hp,
            sorted(list(set(window_hp))),
            plot_hpindex_colors,
            segment=segmentation,
        )
        self.pa.savefig()

        hp_order = sorted(list(hp_index_dict.keys()))
        self.sw.write_segments_llr(
            dense_met_llr,
            segmentation,
            segments_p,
            self.window_met.genomic_coord,
            window_hp,
            self.ps_id,
            hp_index_dict,
            hp_order,
        )

    def error_callback(self, arg):
        start = self.window_met.get_genomic_region()[0]
        end = self.window_met.get_genomic_region()[1]
        print(
            "Warning: Failed for phase set {ps}, window {start}-{"
            "end}".format(ps=self.ps_id, start=start, end=end)
        )
        print(arg)


def segment_phase_set(
    ps: PhaseSet,
    unphased_read_names: List[str],
    sparse_met: net_npc.SparseMethylationMatrixContainer,
    pa: PlotArchiver,
    sw: SegmentsWriterCSV,
    pool: Pool,
    n_segments: int,
):
    ps_id = ps.ps_id

    rn_hp_dict = get_readname_haplotype_dict(ps)
    for read in unphased_read_names:
        rn_hp_dict[read] = 999

    sparse_met = sparse_met.get_submatrix_from_read_names(list(rn_hp_dict.keys()))
    if sparse_met.shape[0] == 0:
        print(
            "Warning: Phase Set %d has %d phased reads but no methylation "
            "calls" % (ps_id, len(ps.read_mappings.keys()))
        )
        return

    winlen = 250
    met_sites = sparse_met.met_matrix.shape[1]
    async_results = []
    for start in range(0, met_sites, winlen):
        print(start)
        window_met = sparse_met.get_submatrix(start, start + winlen)

        callback = SegmentationCallback(sw, pa, window_met, ps_id, ps.chrom)
        set_hp = {rn_hp_dict[read] for read in window_met.read_names}
        if len(set_hp) < 2:
            # No point in segmenting a window that has only reads for a single haplotype
            continue
        ar = pool.apply_async(
            segmentation_worker,
            args=(window_met, rn_hp_dict, n_segments),
            callback=callback.segmentation_callback,
            error_callback=callback.error_callback,
        )
        async_results.append(ar)

    for i, res in enumerate(async_results):
        res.wait(timeout=1800)
        print("%d from %d workers completed" % (i, len(async_results)))'''


def segment(
    sample_h5path: Dict[str, Path],
    genomic_range: Tuple[str, int, int],
    pool: Pool,
    chunk_size: int,
):
    # Open HDF5 files
    sample_h5f: Dict[str, MetH5File] = {
        s: MetH5File(p, "r", chunk_size=chunk_size)
        for s, p in sample_h5path.items()
    }
    # Create dictionary of MethlyationValuesContainer objects for each sample
    sample_met_llrs: Dict[str, MethlyationValuesContainer] = {
        s: f[genomic_range[0]].get_values_in_range(genomic_range[1], genomic_range[2])
        if f[genomic_range[0]] is not None
        else None
        for s, f in sample_h5f.items()
    }

    sparse_met_matrix = create_sparse_matrix_from_samples(sample_met_llrs)
    print(sparse_met_matrix.shape)


def multifile_segmentation(
    h5files: List[str],
    genomic_range: str,
    sample_names: List[str],
    workers: int,
    chunk_size: int,
):

    pool = Pool(workers)
    print(h5files)

    # Parse and validate arguments
    genomic_range = genomic_range.split(":")
    chromosome = genomic_range[0]
    if len(genomic_range) > 1:
        genomic_range = genomic_range[1].split("-")
        try:
            start = int(genomic_range[0].strip())
            end = int(genomic_range[1].strip())
        except ValueError:
            raise ValueError(
                "Invalid format for genomic region. Should be chrom[:start-end]. Start and end must be numeric."
            )
        except IndexError:
            raise ValueError(
                "Invalid format for genomic region. Should be chrom[:start-end]. Both start and end must be provided together and separated by a hyphen."
            )
    else:
        start = 0
        end = sys.maxsize

    # Validate h5 files
    h5files = [Path(f) for f in h5files]
    for f in h5files:
        if not f.exists() or not f.is_file():
            raise ValueError("Invalid HDF5 file, cannot read: \n  %s" % str(f))

    if sample_names is not None:
        if len(sample_names) != len(h5files):
            raise ValueError(
                "If sample names are provided, the number of sample names must be equal to the number of HDF5 files"
            )
    else:
        # Sample names are none, infer from filenames
        sample_names = [f.stem for f in h5files]

    segment(dict(zip(sample_names, h5files)), (chromosome, start, end), pool, chunk_size)



from typing import IO
from io import StringIO

import numpy as np
import pandas as pd


class SegmentsWriterBED:

    def __init__(self, outfile: IO, chrom: str):
        self.outfile = outfile
        self.chrom = chrom
        self.first = True

    def write_segments_llr(
        self,
        llrs: np.ndarray,
        segments: np.ndarray,
        segment_p: np.ndarray,
        genomic_locations: np.ndarray,
        samples: np.ndarray
    ):
        df = pd.DataFrame(columns=["chrom", "start", "end", "num_sites"])
        df = df.astype({"chrom": str, "start": int, "end": int, "num_sites": int})

        for seg in sorted(list(set(segments))):
            seg_pos = np.arange(llrs.shape[1])[segments == seg]
            start = genomic_locations[seg_pos[0]]
            end = genomic_locations[seg_pos[-1]]

            rowval = {
                "chrom": self.chrom,
                "start": start,
                "end": end,
                "num_sites": (segments == seg).sum()
            }

            for i, hp_a in enumerate(hps):
                for hp_b in hps[i + 1 :]:
                    up, pp = compute_differential_methylation(
                        hp_llrs[hp_a], hp_llrs[hp_b]
                    )
                    rowval["unpaired_pval_%s_vs_%s" % (hp_a, hp_b)] = up
                    rowval["paired_pval_%s_vs_%s" % (hp_a, hp_b)] = pp

            rowval.update({"HP%d_p" % hp: hp_p[hp] for hp in hps})

            def format_llrs(x):
                str_list = ["%.2f" % llr for llr in hp_llrs[x][hp_llrs[x] != 0]]
                return ",".join(str_list)

            rowval.update({"HP%d_llrs" % hp: format_llrs(hp) for hp in hps})

            df = df.append(rowval, ignore_index=True)

        df.to_csv(self.outfile, sep="\t", header=True, index=False)
        # self.first = False


class SegmentsReaderCSV:
    """
    Reads a concatenation of CSV files (with potentially differenct columns) and merges
    them into one dataframe, filling in null for missing values. Accepts multiple header
    lines, but requires them to start with the word "chrom" in order to identify the
    header.
    
    This is used to read a CSV file that has been created in chunks by multiple
    concurrent worker threads.
    """
    def __init__(self, *argc):
        self.inputfiles = argc

    def read(self):
        ret = []
        for inputfile in self.inputfiles:
            with open(inputfile) as f:
                cur_string = None
                while True:
                    line = f.readline().strip()
                    if line.startswith("chrom") or not line:
                        # new header line
                        if cur_string is not None:
                            cur_pd = pd.read_csv(
                                StringIO("\n".join(cur_string)), sep="\t"
                            )
                            ret.append(cur_pd)
                        cur_string = []
                    if not line:
                        break
                    cur_string.append(line)
        ret = pd.concat(ret, sort=False)
        ret = ret.sort_values(["chrom", "start", "end"]).reset_index(drop=True)
        return ret


import argparse

import nanoepiseg.main_chunked
import nanoepiseg.main_multi
import nanoepiseg.main_extract_haplotype_ids
import nanoepiseg.main_report


def main():
    parser = argparse.ArgumentParser(
        description="HMM based de-novo segmentation of methylation from Nanopolish methylation calls)"
    )

    parser.add_argument(
        "--chunk_size",
        metavar="chunk_size",
        type=int,
        required=True,
        help="Number of llrs per chunk - for best "
        "performance, should be a multiple of the "
        "chunksize used in creating of the h5 files",
    )

    subparsers = parser.add_subparsers(description="Subcommand: ", dest="subcommand")
    subparsers.required = True

    chunked_args = subparsers.add_parser(
        "single_h5",
        description="Segment a single HDF5 file based on a separate read-to-sample mapping file",
    )
    chunked_args.set_defaults(func=nanoepiseg.main_chunked.chunked_segmentation)

    chunked_args.add_argument(
        "--chunks",
        metavar="chunks",
        type=int,
        nargs="+",
        required=True,
        help="Chunk ids within chromosome in hdf5 file. If more than one hdf5 is "
        "provided, the chunk id is picked from the file with the most chunks, and "
        "the other hdf5 files are mapped based on the genomic coordinates",
    )

    chunked_args.add_argument(
        "--workers",
        metavar="workers",
        type=int,
        help="Number of worker processes",
        default=10,
    )
    chunked_args.add_argument(
        "--h5file",
        metavar="h5file",
        required=True,
        type=str,
        help="H5 file containing Nanopolish methylation calls",
    )
    chunked_args.add_argument(
        "--readgroups",
        metavar="readgroups",
        required=True,
        type=str,
        help="File containing read to readgroup assignment",
    )

    chunked_args.add_argument(
        "--include_nogroup",
        action="store_true",
        default=False,
        dest="include_nogroup",
        help="Include reads with group -1 in all windows",
    )

    multi_h5_args = subparsers.add_parser(
        "multi_h5",
        description="Segment multiple H5 files, where each h5 file represents one sample",
    )
    multi_h5_args.set_defaults(func=nanoepiseg.main_multi.multifile_segmentation)

    multi_h5_args.add_argument(
        "--genomic_range",
        metavar="genomic_range",
        type=str,
        required=True,
        help="Genomic range in format chr[:start-end]",
    )
    multi_h5_args.add_argument(
        "--workers",
        metavar="workers",
        type=int,
        help="Number of worker processes",
        default=10,
    )
    multi_h5_args.add_argument(
        "--h5files",
        metavar="h5files",
        nargs="+",
        type=str,
        required=True,
        help="H5 files containing Nanopolish methylation calls, one file per sample",
    )
    multi_h5_args.add_argument(
        "--sample_names",
        metavar="sample_names",
        nargs="*",
        type=str,
        help="Sample names (one per h5 file). If not provided, will be inferred from the filenames",
    )

    extract_haplotype_ids_args = subparsers.add_parser(
        "extract_haplotype_ids",
        description="Extract phase set and haplotype id from bam file",
    )
    extract_haplotype_ids_args.set_defaults(
        func=nanoepiseg.main_extract_haplotype_ids.extract_haplotype_ids
    )

    extract_haplotype_ids_args.add_argument(
        "--bam",
        metavar="bam",
        type=str,
        required=True,
        help="BAM File containing read group annotation",
    )

    extract_haplotype_ids_args.add_argument(
        "--output",
        metavar="output",
        type=str,
        required=True,
        help="Output tsv file",
    )
    extract_haplotype_ids_args.add_argument(
        "--include_unphased",
        action="store_true",
        default=False,
        dest="include_unphased",
        help="Also include reads that have no HP tag (will be stored as -1/-1 ps/hp)",
    )
    extract_haplotype_ids_args.add_argument(
        "--chroms",
        metavar="chromosomes",
        type=str,
        required=False,
        nargs="*",
        help="Chromosomes to consider (default is all)",
    )

    report_args = subparsers.add_parser("report",
        description="Plot methylation profile for regions", )
    report_args.set_defaults(func=nanoepiseg.main_report.report)

    report_args.add_argument("--regions_tsv_fn", metavar="regions_tsv_fn", type=str,
        required=True, help="TSV output of pycometh report", )
    report_args.add_argument("--gene_expression_file", metavar="regions_tsv_fn", type=str,
        required=True, help="TSV output of pycometh report", )
    report_args.add_argument("--h5_fns", metavar="h5_fns", type=str, nargs="+",
        required=True,
        help="H5 files containing Nanopolish methylation calls, one per sample", )
    report_args.add_argument("--output_pdf_fn", metavar="output_pdf_fn", type=str,
        required=True,
        help="PDF output file", )

    args = parser.parse_args()
    args_dict = vars(args)
    # Remove arguments that the subcommand doesn't take
    subcommand = args.func
    del args_dict["subcommand"]
    del args_dict["func"]

    subcommand(**args_dict)



import math
from abc import ABC, abstractmethod
from typing import Dict

import numpy as np


class EmissionLikelihoodFunction(ABC):
    """
    Abstract class defining the interface for emission likelihood functions
    The types for parameters is not defined here (only that it's stored as Dict) so that
    the concrete implementation of the likelihood function may define it's own
    domain of parameters, as long as the update_params function takes the same structure
    of parameters as the get_params and get_cluster_params method return
    """
    @abstractmethod
    def update_params(self, params: Dict):
        """
        :param params: Dictionary where the key is the cluster id and the value contains
        the parameter(s)
        :type params: Dict
        :return:
        """
        pass

    @abstractmethod
    def likelihood(self, segment_index: int, observations: np.array):
        pass

    @abstractmethod
    def get_cluster_params(self, cluster: int):
        pass

    @abstractmethod
    def minimization_objective(self, observations: np.array,
                               posterior: np.array):
        pass

    @abstractmethod
    def get_param_bounds(self):
        pass


class BernoulliPosterior(EmissionLikelihoodFunction):
    """
    This class models a Bernoulli likelihood with uncertain observations.
    Observations are provided as probabilities p(a|S), and the likelihood
    is parameterized with bernoulli-likelihoods p(a|mu) for each segment.
    Optionally, a gamma prior p(mu) can be defined as well.
    The likelihood is modeled as:
    L = (1-p(a|S)) * (1-mu) / (1-p(a)) + p(a|S) * mu / p(a)
    """

    def __init__(self, number_of_clusters, number_of_segments,
                 prior_a: float = None, eps=np.exp(-512),
                 initial_segment_p: np.array = None):
        # Just for type hinting and to get rid of warnings
        self.segment_p: np.array = None
        self.segment_prior: np.array = None
        self.prior_lognormfactor: float = 0

        self.eps: float = eps
        self.prior_a: float = prior_a

        if initial_segment_p is None:
            self.segment_p = np.zeros((number_of_clusters, number_of_segments))
            self.segment_p[:, ::2] = np.log(1 / 5)
            self.segment_p[:, 1::2] = np.log(4 / 5)
        else:
            if not (initial_segment_p.shape[0] == number_of_clusters and
                    initial_segment_p.shape[0] == number_of_segments):
                raise ValueError('Initial parameters must be of shape ('
                                 'number_of_clusters, number_of_segments)')
            self.segment_p = initial_segment_p.copy()

        if self.prior_a is not None:
            # Precompute the normfactor (in log space) of the prior gamma
            # distribution
            self.prior_lognormfactor = np.log(
                math.gamma(2 * prior_a) / (math.gamma(prior_a) ** 2))

    def update_prior(self):
        """
        Updates the prior beta distribution (in log space) based on
        self.segment_p and self.prior_a
        """
        self.segment_prior = self.segment_p * (self.prior_a - 1)
        self.segment_prior += np.log(1 - np.exp(self.segment_p) + self.eps) * (
                self.prior_a - 1)
        self.segment_prior += self.prior_lognormfactor

    def update_params(self, segment_p_list: Dict[int, np.ndarray]):
        """
        Updates bernoulli parameters
        :param segment_p_list: bernoulli parameters. A list with one numpy array
        per cluster
        :return: difference between old and new parameters to assess convergence
        """
        maxdiff = 0
        for sample in segment_p_list.keys():
            diff = np.max(np.abs(np.exp(self.segment_p[sample, :]) - np.exp(
                segment_p_list[sample])))
            maxdiff = max(diff, maxdiff)
            self.segment_p[sample, :] = segment_p_list[sample]

        if self.prior_a is not None:
            self.update_prior()
        return maxdiff

    def likelihood(self, segment_index: int, observations: np.array,
                   observations_cluster_assignment: np.array):
        p = self.segment_p[observations_cluster_assignment, :]
        idx = (observations != -1)


        ret_a = (np.log(-np.expm1(p[idx, segment_index]) + self.eps))
        ret_a += np.log1p(-observations[idx]) + np.log(0.5)

        ret_b = p[idx, segment_index]
        ret_b += np.log(observations[idx] + self.eps) + np.log(0.5)

        ret = np.logaddexp(ret_a, ret_b)
        if self.segment_prior is not None:
            ret += self.segment_prior[observations_cluster_assignment, :][
                idx, segment_index]
        return ret.sum()

    def minimization_objective(self, observations: np.array,
                               posterior_exp: np.array):
        """
        Returns a curried function that only takes the candidate parameters
        mu and returns a minimization object (in this case the total
        likelihood p(S|mu',psi) given S and posteriors p(psi|mu))
        :param observations: observations numpy array
        :param posterior_exp: posterior of segmentation as estimated by the hmm
        - must be in linear space (not log space)
        :return: a function that takes parameters mu and returns likelihood
        """

        def curried_objective(x):
            # x is in log-space but these computations are easier in
            # lin-space
            # since there are a lot of additions going on
            # x = np.exp(x)
            ls = np.zeros(self.segment_p.shape[1])
            ps = np.zeros(self.segment_p.shape[1])
            for r in range(observations.shape[0]):
                o = observations[r, :]
                idx = o != -1
                o = o[idx]
                pki = posterior_exp[idx, :]

                l_a = np.outer((1 - o), (1 - x)) / 2
                l_b = np.outer(o, x) / 2
                l = np.log(l_a + l_b + self.eps)
                ls += (l * pki).sum(axis=0)
                ps += pki.sum(axis=0)

            ret = ls / (ps + self.eps)
            ret = ret.sum()
            return -ret

        return curried_objective

    def get_cluster_params(self, cluster):
        """
        Returns parameters for optimization for one cluster
        :param cluster: cluster index
        :return: parameters in linear space
        """
        # Clipping is needed due to a scipy bug
        return np.clip(np.exp(self.segment_p[cluster, :]), 0.02, 0.98)

    def get_params(self):
        # Clipping is needed due to a scipy bug
        return np.clip(np.exp(self.segment_p), 0.02, 0.98)

    def get_param_bounds(self):
        """
        :return: parameters bounds for optimization (interval [0,1])
        """
        return [(0.01, 0.99)] * self.segment_p.shape[1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup

# Long description from README file
with open("README.md", "r") as fh:
    long_description = fh.read()

# Collect info in a dictionary for setup.py
setup(
    name="nanoepiseg",
    description="Methylome segmentation algorithm using a changepoint detection HMM",
    version="0.0.2.dev1",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/snajder-r/nanoepiseg",
    author="Rene Snajder",
    author_email="r.snajder@dkfz-heidelberg.de",
    license="MIT",
    python_requires=">=3.7",
    classifiers=["Development Status :: 3 - Alpha", "Intended Audience :: Science/Research", "Topic :: Scientific/Engineering :: Bio-Informatics", "License :: OSI Approved :: MIT License", "Programming Language :: Python :: 3"],
    install_requires=["numpy>=1.19.2", "scipy==1.4.1", "pandas>=1.1.3", "meth5>=0.2.6"],
    packages=["nanoepiseg"],
    package_dir={"nanoepiseg": "nanoepiseg"},
    entry_points={"console_scripts": ["nanoepiseg=nanoepiseg.__main__:main"]},
)



import numpy as np


def cleanup_segmentation(segment_p: np.ndarray, segments: np.ndarray,
                         min_length: int = 5,
                         min_parameter_diff=0.1) -> np.ndarray:
    """
    Cleans up a segmentation by merging segments that are too close or too
    similar in their parameter space.
    :param segment_p: segment parameters. Shape (C,M) where C is the number
    of clusters and M is the number of segments
    :param segments: segmentation as returned from the viterbi or MAP algorithm.
    Shape is (N) where N is the number of genomic indices (i.e. CpG sites).
    :param min_length: minimum number of genomic loci per segment
    :param min_parameter_diff: minimum difference in parameter values between
    neighboring segments
    :return: a new segmentation of the same shape as segments
    """
    new_segments = segments.copy()
    lastx = 0
    for segment in sorted(list(set(segments))):
        if len(set(new_segments)) <= 1:
            # No need to go on if it's all just one segment
            break
        length = (new_segments == segment).sum()
        if segment == new_segments[-1]:
            cadidate_replace = new_segments[new_segments != segment][-1]
        else:
            cadidate_replace = segment + 1
        absdif = np.abs(
            segment_p[:, segment] - segment_p[:, cadidate_replace]).max()
        if length < min_length or absdif < min_parameter_diff:
            new_segments[new_segments == segment] = cadidate_replace

    return np.array(new_segments)
